/*
    Copyright 2005-2008 Intel Corporation.  All Rights Reserved.

    This file is part of Threading Building Blocks.

    Threading Building Blocks is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    Threading Building Blocks is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Threading Building Blocks; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    As a special exception, you may use this file as part of a free software
    library without restriction.  Specifically, if other files instantiate
    templates or use macros or inline functions from this file, or you compile
    this file and link it with other files to produce an executable, this
    file does not by itself cause the resulting executable to be covered by
    the GNU General Public License.  This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.
*/

#ifndef __TBB_machine_H
#error Do not include this file directly; include tbb_machine.h instead
#endif

#include "linux_common.h"

#define __TBB_WORDSIZE 4
#define __TBB_BIG_ENDIAN 0

#define __TBB_PLATFORM_ORDERS_STORES 1
#define __TBB_PLATFORM_ORDERS_LOADS 1
static inline void __TBB_fence_for_release () { __asm__ __volatile__("": : :"memory"); }
static inline void __TBB_fence_for_acquire () { __asm__ __volatile__("": : :"memory"); }
static inline void __TBB_fence_for_compiler() { __asm__ __volatile__("": : :"memory"); }

namespace tbb { namespace internal {

// __TBB_CompareAndSwap

    #define __TBB_M(S,T,X)                                               \
    template<memory_semantics M> struct __TBB_CompareAndSwap<S,T,M> {    \
        static inline T op( volatile void *ptr, T value, T comparand ) { \
            __TBB_fence_guard<true,true,true,M> anonymous;               \
            T result;                                                    \
            __asm__ __volatile__(                                        \
                "lock\ncmpxchg" X " %2,%1"                               \
                : "=a"(result), "=m"(*(T *)ptr)                          \
                : "q"(value), "0"(comparand), "m"(*(T *)ptr)             \
                : );                                                     \
            return result;                                               \
        }                                                                \
    };

    __TBB_M(1,int8_t ,"b")
    __TBB_M(2,int16_t,"w")
    __TBB_M(4,int32_t,"l")

    #undef __TBB_M

    // TODO: make run-time decision to use 8-byte CMPXCHG instead, if available?
    template<memory_semantics M> struct __TBB_CompareAndSwap<8,int64_t,M> {
        static inline int64_t op( volatile void *ptr, int64_t value, int64_t comparand ) {
            __TBB_fence_guard<true,true,true,M> anonymous;
            int64_t result;
            /* EBX register saved for compliancy with position-independent code (PIC) rules on IA32 */
            __asm__ __volatile__ (
                "pushl %%ebx\n\t"
                "movl  (%%ecx),%%ebx\n\t"
                "movl  4(%%ecx),%%ecx\n\t"
                "lock\ncmpxchg8b (%2)\n\t"
                "popl  %%ebx"
                : "=A"(result), "=m"(*(int64_t *)ptr)
                : "S"(ptr),
                  "a"(((int32_t*)&comparand)[0]),
                  "d"(((int32_t*)&comparand)[1]),
                  "c"(&value)
                : );
            return result;
        }
    };

// __TBB_Load, __TBB_Op

    #define __TBB_MM(S,T,X,__TBB_op_op,mnemonic)                              \
    template<memory_semantics M> struct __TBB_Op<S,T,M,__TBB_op_op> {         \
        static inline void op( volatile void *ptr, T value) {                 \
            __TBB_fence_guard<true,true,false,M> anonymous;                   \
            __asm__ __volatile__("lock\n" mnemonic X " %1,%0"                 \
                : "=m"(*(T *)ptr) : "r"(value) : );                           \
        }                                                                     \
    };

    #define __TBB_M(S,T,X)                                                    \
    template<memory_semantics M> struct __TBB_Load<S,T,M> {                   \
        static inline T op (const volatile void *ptr) {                       \
            __TBB_fence_guard<false,false,true,M> anonymous;                  \
            /* TODO: use assembler instead of relying on the compiler? */     \
            return *((volatile T*)(ptr));                                     \
        }                                                                     \
    };                                                                        \
    template<memory_semantics M> struct __TBB_Op<S,T,M,__TBB_op_store> {      \
        static inline void op (volatile void *ptr, T value) {                 \
            __TBB_fence_guard<true,false,false,M> anonymous;                  \
            /* TODO: use assembler instead of relying on the compiler? */     \
            *((volatile T*)(ptr)) = value;                                    \
        }                                                                     \
    };                                                                        \
    __TBB_MM(S,T,X,__TBB_op_and,"and")                                        \
    __TBB_MM(S,T,X,__TBB_op_or ,"or" )                                        \
    __TBB_MM(S,T,X,__TBB_op_xor,"xor")

    __TBB_M(1,int8_t ,"b")
    __TBB_M(2,int16_t,"w")
    __TBB_M(4,int32_t,"l")

    #undef __TBB_M

    #undef __TBB_MM

    template<memory_semantics M> struct __TBB_Load<8,int64_t,M> {
        static inline int64_t op (const volatile void *ptr) {
            __TBB_fence_guard<false,false,true,M> anonymous;
            int64_t result;
            __asm__ __volatile__ ( "fildq %1\n\t"
                                   "fistpq %0" :  "=m"(result) : "m"(*(uint64_t *)ptr), "m"(result) : );
            return result;
        }
    };
    template<memory_semantics M> struct __TBB_Op<8,int64_t,M,__TBB_op_store> {
        static inline void op (volatile void *ptr, int64_t value) {
            __TBB_fence_guard<true,false,false,M> anonymous;
            __asm__ __volatile__ ( "fildq %1\n\t"
                                   "fistpq (%2)" :  "=m"(*(int64_t *)ptr) : "m"(value), "r"(ptr) : );
        }
    };

// __TBB_FetchOp

    #define __TBB_MM(S,T,X,__TBB_op_op,mnemonic)                               \
        template<memory_semantics M> struct __TBB_FetchOp<S,T,M,__TBB_op_op> { \
            static inline T op (volatile void *ptr, T arg) {                   \
                __TBB_fence_guard<true,true,true,M> anonymous;                 \
                T result;                                                      \
                __asm__ __volatile__(                                          \
                    mnemonic X " %0,%1"                                        \
                    : "=r"(result), "=m"(*(T *)ptr)                            \
                    : "0"(arg), "m"(*(T *)ptr)                                 \
                    : );                                                       \
               return result;                                                  \
            }                                                                  \
        };

    #define __TBB_M(S,T,X)                                               \
        __TBB_MM(S,T,X,__TBB_op_store,      "xchg") /*lock is implicit*/ \
        __TBB_MM(S,T,X,__TBB_op_add  ,"lock\nxadd")
        // there are no "Exchange and Logical {AND,OR,XOR}" operations

    __TBB_M(1,int8_t ,"b")
    __TBB_M(2,int16_t,"w")
    __TBB_M(4,int32_t,"l")
    // TODO: anything for 8 bytes?

    #undef __TBB_M

    #undef __TBB_MM

} } // tbb::internal

static inline int32_t __TBB_machine_lg( uint32_t x ) {
    int32_t j;
    __asm__ ("bsr %1,%0" : "=r"(j) : "r"(x));
    return j;
}

static inline void __TBB_machine_pause( int32_t delay ) {
    for (int32_t i = 0; i < delay; i++) {
       __asm__ __volatile__("pause;"); // TODO: what does the semicolon do, if anything?
    }
    return;
}   

// Machine specific atomic operations

// TODO: instead eliminate __TBB_machine_cmpswp... etc., and only redefine W in terms of 4?
// or do these have to be macros?

// used in spin_rw_mutex.cpp, TODO: what memory semantics?
#define __TBB_CompareAndSwapW(P,V,C) (tbb::internal::__TBB_CompareAndSwap<4,int32_t,tbb::rel_acq>::op(P,V,C))

// used in task.cpp, TODO: what memory semantics?
#define __TBB_FetchAndStoreW(P,V) (tbb::internal::__TBB_FetchOp<4,int32_t,tbb::rel_acq,tbb::internal::__TBB_op_store>::op(P,V))

// used in spin_rw_mutex.cpp, TODO: what memory semantics?
#define __TBB_FetchAndAddW(P,V) (tbb::internal::__TBB_FetchOp<4,int32_t,tbb::rel_acq,tbb::internal::__TBB_op_add>::op(P,V))

// Definition of other functions
#define __TBB_Pause(V) __TBB_machine_pause(V)
#define __TBB_Log2(V)  __TBB_machine_lg(V)

// Definition of Lock functions
#undef __TBB_TryLockByte
#undef __TBB_LockByte

#define __TBB_cpuid
static inline void __TBB_x86_cpuid( int32_t buffer[4], int32_t mode ) {
    // EBX register saved for compliancy with position-independent code (PIC) rules on IA32
    __asm__ ("pushl %%ebx\n\t"
             "cpuid\n\t"
             "movl  %%ebx,%1\n\t"
             "popl  %%ebx"
                    : "=a"(buffer[0]),
                      "=S"(buffer[1]),
                      "=c"(buffer[2]),
                      "=d"(buffer[3])
                    : "0"(mode)
                    : "memory" );
}

