/*
    Copyright 2005-2008 Intel Corporation.  All Rights Reserved.

    This file is part of Threading Building Blocks.

    Threading Building Blocks is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    Threading Building Blocks is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Threading Building Blocks; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    As a special exception, you may use this file as part of a free software
    library without restriction.  Specifically, if other files instantiate
    templates or use macros or inline functions from this file, or you compile
    this file and link it with other files to produce an executable, this
    file does not by itself cause the resulting executable to be covered by
    the GNU General Public License.  This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.
*/

#ifndef __TBB_machine_H
#error Do not include this file directly; include tbb_machine.h instead
#endif

#include "linux_common.h"
#include <ia64intrin.h>

#define __TBB_WORDSIZE 8
#define __TBB_BIG_ENDIAN 0

TODO: good look at what has changed (this is just architectural so far)
TODO: add __TBB_Load and __TBB_Op<__TBB_op_store>

#define __TBB_PLATFORM_ORDERS_STORES TODO
#define __TBB_PLATFORM_ORDERS_LOADS TODO
#ifdef __INTEL_COMPILER
TODO: really? reflect in comments...
static inline void __TBB_fence_for_release () { }
static inline void __TBB_fence_for_acquire () { }
#else
/* Even though GCC imbues volatile loads with acquire semantics, 
   it sometimes moves loads over the acquire fence.  The
   fences defined here stop such incorrect code motion. */
static inline void __TBB_fence_for_release () { __asm__ __volatile__("": : :"memory"); }
static inline void __TBB_fence_for_acquire () { __asm__ __volatile__("": : :"memory"); }
#endif
static inline void __TBB_fence_for_compiler() { __asm__ __volatile__("": : :"memory"); }

// Most of the functions will be in a .s file

// __TBB_CompareAndSwap, __TBB_FetchOp

    // TODO: is the __TBB_full_fence (from before the change) now seq_cst/ordered or just rel_acq?
    // (the first is currently assumed); rename to what is used at C++ level and eliminate separate parameter

    #define __TBB_M_STM(S,T,M1,M2)                                                     \
        extern "C" {                                                                   \
            T __TBB_machine_cmpswp    ##S##M2(volatile void *ptr, T arg, T comparand); \
            T __TBB_machine_fetchstore##S##M2(volatile void *ptr, T arg             ); \
            T __TBB_machine_fetchadd  ##S##M2(volatile void *ptr, T arg             ); \
        }                                                                              \
        template<> struct __TBB_CompareAndSwap<S,T,M1> {                               \
            static inline T op( volatile void *ptr, T arg, T comparand ) {             \
                return __TBB_machine_cmpswp##S##M2( ptr, arg, comparand );             \
            }                                                                          \
        };                                                                             \
        template<> struct __TBB_FetchOp<S,T,M1,__TBB_op_store> {                       \
            static inline T op( volatile void *ptr, T arg ) {                          \
                return __TBB_machine_fetchstore##S##M2( ptr, arg );                    \
            }                                                                          \
        };                                                                             \
        template<> struct __TBB_FetchOp<S,T,M1,__TBB_op_add> {                         \
            static inline T op( volatile void *ptr, T arg ) {                          \
                return __TBB_machine_fetchadd##S##M2( ptr, arg );                      \
            }                                                                          \
        };

    #define __TBB_M_M(M1,M2)         \
        __TBB_M_STM(1,int8_t ,M1,M2) \
        __TBB_M_STM(2,int16_t,M1,M2) \
        __TBB_M_STM(4,int32_t,M1,M2) \
        __TBB_M_STM(8,int64_t,M1,M2) \

    __TBB_M_M(ordered,__TBB_full_fence)
    __TBB_M_M(acquire,acquire         )
    __TBB_M_M(release,release         )

    #undef __TBB_M_M

    #undef __TBB_M_STM

extern "C" {
    int64_t __TBB_machine_lg(uint64_t value);
    void __TBB_machine_pause(int32_t delay);
    bool __TBB_machine_trylockbyte( volatile unsigned char &ptr );
    int64_t __TBB_machine_lockbyte( volatile unsigned char &ptr );
}

// Special atomic functions
#define __TBB_CompareAndSwapW(P,V,C)   __TBB_CompareAndSwap8(P,V,C)
#define __TBB_FetchAndStoreW(P,V)      __TBB_FetchAndStore8(P,V)
#define __TBB_FetchAndAddW(P,V)        __TBB_FetchAndAdd8(P,V)
#define __TBB_FetchAndAddWrelease(P,V) __TBB_FetchAndAdd8release(P,V)

// Not needed
#undef __TBB_Store8
#undef __TBB_Load8

// Definition of Lock functions
#define __TBB_TryLockByte(P) __TBB_machine_trylockbyte(P)
#define __TBB_LockByte(P)    __TBB_machine_lockbyte(P)

// Definition of other utility functions
#define __TBB_Pause(V) __TBB_machine_pause(V)
#define __TBB_Log2(V)  __TBB_machine_lg(V)

